#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <io.h>
#include <dos.h>
#include <sys\stat.h>
#include <sys\types.h>
#include "midifile.h"

#define MAXNOTES 4096
#define MAXCHANNELS 9
#define NUMINSTS 256

FILE *fp;

typedef struct
{
	short freq, track, volume;
	short event, paramanter;
} Note_T;

typedef struct
{
    	char filler[2];
    	char filler1[2];
    	char filler2[2];
    	char filler3[2];
    	char filler4[2];
    	char filler5;
	short mt32inst, geninst;
} INSTDATA;

short numnotes, trinst[MAXCHANNELS], songtempo;
short songspeed, skiptime, sysexsize;
char sngsig[4], sysex[64];
Note_T note[MAXNOTES];
INSTDATA inst[NUMINSTS];

myputc(char c) { return(putc(c,fp));}

int writesong(int track)
{
	long i, j;
	char data[2], ch;

	j = 1000 * 60 / songtempo;
        mf_write_tempo((long)j*1000);
	for(i=1;i<MAXCHANNELS;i++)
	{
		data[0] = (inst[trinst[i]].geninst&0x7f);
                if(!mf_write_midi_event(0,program_chng,i,data,1))
			return(-1);
	}
	for(i=0;i<numnotes;i++)
	{
		switch (note[i].event)
		{
			case 2:	
			data[0] = 0x7b;
			data[1] = 0;
                	if(!mf_write_midi_event(0,control_change,note[i].paramanter,data,2))
				return(-1);
			break;
			case 3:
			songspeed = note[i].paramanter;
			break;
			case 4:
			data[0] = 0x79;
			data[1] = 0;
                	if(!mf_write_midi_event(0,control_change,note[i].paramanter,data,2))
				return(-1);
			break;
			case 5:
			data[0] = 0x40;
			data[1] = note[i].paramanter;
                	if(!mf_write_midi_event(0,control_change,note[i].track,data,2))
				return(-1);
			break;
			case 7:
			sysex[sysexsize++] = note[i].paramanter;
			break;
			case 8:
			sysex[sysexsize++] = 0xf7;
                	if(!mf_write_midi_event(0,system_exclusive,0,sysex,sysexsize))
				return(-1);
			sysexsize = 0;
			break;
		}
		if (note[i].freq != 0)
		{
			if (inst[trinst[note[i].track]].geninst < 127)
			{
				ch = note[i].track;
			}
			else
			{
				ch = 9;
			}
			data[0] = 0x7b;
			data[1] = 0;	
                	if(!mf_write_midi_event(songspeed*4+skiptime,control_change,ch,data,2))
				return(-1);
			if (inst[trinst[note[i].track]].geninst < 127)
			{
				ch = note[i].track;
				if (note[i].freq < 90)
					note[i].freq += 35;
				data[0] = note[i].freq;
				data[1] = note[i].volume*2;
			}
			else
			{
				ch = 9;
				data[0] = inst[trinst[note[i].track]].geninst-100;
				data[1] = note[i].volume*2;
			}
                	if(!mf_write_midi_event(0,note_on,ch,data,2))
				return(-1);
			skiptime = 0;
		}
		else
			skiptime += songspeed*4;
	}
	return(1);
}

void loadinst(void)
{
        long fil;

        fil = loaditem("INSTS.DAT");
        if (fil == -1) return;
        readitem(fil, &inst[0], sizeof(INSTDATA)*256);
        closeitem(fil);
}

int sng2midi_main(char *name, char *srcname)
{
	unsigned long i;
	long fil;

	loadinst();
	fil = loaditem(srcname);
		
	readitem(fil, &sngsig, 4);
	if (sngsig[0] != 'S' && sngsig[1] != 'N' && sngsig[2] != 'G'
		&& sngsig[3] != 0x1a)
	{
		return(-1);
	}
	readitem(fil, &songtempo, 2);
	readitem(fil, &numnotes, 2);
	readitem(fil, &trinst[0], 2*MAXCHANNELS);
	readitem(fil, &note[0], sizeof(Note_T)*numnotes);
	closeitem(fil);
	songspeed = 10;
	printf("There are %d notes in the song\n", numnotes);
	printf("Writing MIDI file....\n");
	if ((fp = fopen(name,"wb")) == 0)
		return(-1);
        Mf_putc = myputc;
        Mf_writetrack = writesong;

        /* write a single track */
        mfwrite(0,1,48,fp);
	fclose(fp);
	printf("Song converted.\n");
	return(0);
}
